Program pepseq;

uses CRT;

type s1   = string[1];
     s5   = string[5];
     s6   = string[6];
     s8   = string[8];
     s12  = string[12];
     s50  = string[50];

const H   =  1.00782;
      OH  = 17.0027;
      H2O = 18.01052;
      CO  = 27.99491;
      NH  = 15.01090;

var npeak : integer;                   {number of peaks in sequence}
    nspec : integer;                   {number of peaks in spectrrum}
    cpeak : array[1..600] of real;     {calculated peaks in sequence}
    speak : array[1..200] of real;     {peaks found in spectrum}
    amiw  : array[1..255] of real;     {aminoacid weight}
    amic  : array[1..255] of byte;     {aminoacid code}
    amix  : array[1..255] of byte;     {quantity of aminoacids}
    amir  : byte;                      {number of aminoacids read}
    score : real;                      {score for the fit}
    molw  : real;                      {target molw}
    nend  : real;                      {N terminus}
    cend  : real;                      {C terminus}
    molc  : real;                      {calculated molw for sequence}
    mina  : byte;                      {minimum amino acids in sequence}
    maxa  : byte;                      {maximum amino acids in sequence}
    maxs  : real;                      {maximum score for the sequence}
    maxp  : real;                      {maximum score per peak}
    maxh  : longint;                   {maximum hits}
    tole  : real;                      {tolerance for calculated molw}
    tolp  : real;                      {tolerance for measured peaks}
    hitp  : byte;                      {peaks recognized}
    minp  : byte;                      {minimum peaks to be hit}
    ions  : s6;                        {fragment ions to be calculated}
    titl  : s50;                       {title for report}
    outf  : text;                      {output file}
    repo  : s8;                        {output format}
    datf  : s12;                       {file containg amino acid data}
    inpf  : s12;                       {file containing input}
    check : longint;                   {sequences checked}
    hit   : longint;                   {number of hits}
    temp  : real;
    rot   : byte;

{***************}
Procedure stopit;                      {abort program}
{***************}

begin
  writeln(outf);
  writeln(outf,'CHECKED :',check:10);
  writeln(outf,'HITS    :',hit:10);
  if keypressed then writeln(outf,'*** Program halted ***');
  close(outf);
  gotoxy(1,10);
  if keypressed or (hit>=maxh) then writeln(#7,'*** Program halted ***')
  else writeln(#7,'*** Program ended ***');
  halt;
end;
    
{*****************************}
Procedure makerep(var pep:s50);        {make report}
{*****************************}
var i,j : byte;

begin
  inc(hit);
  gotoxy(21,2);write(hit:10);
  for i:=1 to ord(repo[0]) do
  case upcase(char(repo[i])) of
    '#' : write(outf,hit:5);
    'Q' : write(outf,pep:maxa+1);
    'M' : write(outf,molc:10:4);
    'D' : write(outf,molc-molw:10:4);
    'S' : write(outf,score:8:2);
    '/' : if hitp>0 then write(outf,score/hitp:8:2) else write(outf,'  *':8);
    'P' : write(outf,hitp:4);
    'C' : for j:=1 to npeak do
          begin
            write(outf,cpeak[j]:8:2);
          end;
  end;
  writeln(outf);
  if keypressed or (hit>=maxh) then stopit;
end;

{****************}
Procedure rephead;                     {make report header}
{****************}

var i : byte;

begin
writeln(outf,titl);writeln(outf);
writeln(outf,'MAXA = ',maxa:10,  '     MOLW = ',molw:10:4,  '     DATA = ',datf:14);
writeln(outf,'MINA = ',mina:10,  '     MOLH = ',molw+H:10:4,'     INPT = ',inpf:14);
writeln(outf,'MINP = ',minp:10,  '     TOLE = ',tole:10:4,  '     REPO = ',repo:14);
writeln(outf,'MAXS = ',maxs:10:4,'     NEND = ',nend:10:4,  '     IONS = ',ions:14);
writeln(outf,'MAXP = ',maxp:10:4,'     CEND = ',cend:10:4);
writeln(outf,'TOLP = ',tolp:10:4,'     MAXH = ',maxh:10);writeln(outf);
for i:=1 to amir do
begin
  write(outf,'  ',chr(amic[i]),':',amiw[amic[i]]:10:4);
  if (i mod 5=0) or (i=amir) then writeln(outf);
end;
writeln(outf);
for i:=1 to nspec do
begin
  write(outf,i:3,':',speak[i]:10:4);
  if (i mod 5=0) or (i=nspec) then writeln(outf);
end;
writeln(outf);
for i:=1 to ord(repo[0]) do
case upcase(char(repo[i])) of
  '#' : write(outf,' HIT#':5);
  'Q' : write(outf,'':maxa+1);
  'M' : write(outf,'   MOLW':10);
  'D' : write(outf,'   DEV':10);
  'S' : write(outf,'   SC':8);
  '/' : write(outf,'  SC/P':8);
  'P' : write(outf,'   P':4);
  'C' : write(outf,'':npeak*8);
end;
writeln(outf);writeln(outf);
end;

{*********************************}
Function calcs(var pk:real) : real;    {calculate score}
{*********************************}

var i,recpk : byte;
    mins    : real;

begin
  mins:=9999;
  for i:=1 to npeak do
  begin
    if abs(cpeak[i]-pk)<mins then
    begin
      mins:=abs(cpeak[i]-pk);
      recpk:=i;
    end;
  end;
  if mins<tolp then
  begin
    calcs:=mins;
    cpeak[recpk]:=-cpeak[recpk];
    inc(hitp);
  end
  else calcs:=0;
end;

{******************************}
Procedure makepeak(var pep:s50);       {calculate theoretical peaks}
{******************************}

var i,j,seqlen,posa,posb,posc,posx,posy,posz : word;

begin
  seqlen:=ord(pep[0]);
  posa:=pos('A',ions);
  posb:=pos('B',ions);
  posc:=pos('C',ions);
  posx:=pos('X',ions);
  posy:=pos('Y',ions);
  posz:=pos('Z',ions);
  for i:=1 to seqlen-1 do
  begin
    if posa<>0 then                                         {a-series}
    begin
      j:=(posa-1)*(seqlen-1);
      if i=1 then cpeak[j+i]:=nend+amiw[ord(pep[i])]-CO
      else cpeak[j+i]:=cpeak[j+i-1]+amiw[ord(pep[i])];
    end;
    if posb<>0 then                                         {b-series}
    begin
      j:=(posb-1)*(seqlen-1);
      if i=1 then cpeak[j+i]:=nend+amiw[ord(pep[i])]
      else cpeak[j+i]:=cpeak[j+i-1]+amiw[ord(pep[i])];
    end;
    if posc<>0 then                                         {c-series}
    begin
      j:=(posc-1)*(seqlen-1);
      if i=1 then cpeak[j+i]:=nend+NH+2*H+amiw[ord(pep[i])]
      else cpeak[j+i]:=cpeak[j+i-1]+amiw[ord(pep[i])];
    end;
    if posx<>0 then                                         {x-series}
    begin
      j:=posx*(seqlen-1);
      if i=1 then cpeak[j]:=cend+CO+amiw[ord(pep[seqlen])]
      else cpeak[j-i+1]:=cpeak[j-i+2]+amiw[ord(pep[seqlen-i+1])];
    end;
    if posy<>0 then                                         {y-series}
    begin
      j:=posy*(seqlen-1);
      if i=1 then cpeak[j]:=cend+2*H+amiw[ord(pep[seqlen])]
      else cpeak[j-i+1]:=cpeak[j-i+2]+amiw[ord(pep[seqlen-i+1])];
    end;
    if posz<>0 then                                         {z-series}
    begin
      j:=posz*(seqlen-1);
      if i=1 then cpeak[j]:=cend+amiw[ord(pep[seqlen])]-NH
      else cpeak[j-i+1]:=cpeak[j-i+2]+amiw[ord(pep[seqlen-i+1])];
    end;
  end;
  npeak:=0;
  if posa<>0 then inc(npeak,seqlen-1);
  if posb<>0 then inc(npeak,seqlen-1);
  if posc<>0 then inc(npeak,seqlen-1);
  if posx<>0 then inc(npeak,seqlen-1);
  if posy<>0 then inc(npeak,seqlen-1);
  if posz<>0 then inc(npeak,seqlen-1);
end;


{*****************************}
Procedure perm(pep:s50;x:byte);        {find all possible permutations (n!)}
{*****************************}

var i,j,k : byte;
    ch    : char;
    
begin
  if (x=2) and (pep[1]=pep[2]) then dec(x);
  if x>1 then for i:=1 to x do
  begin
    rot:=122-rot;
    gotoxy(32,2);write(char(rot));
    if i>1 then
    begin
      ch:=pep[x];
      for j:=x downto 2 do
      begin
        pep[j]:=pep[j-1];
      end;
      pep[1]:=ch;
    end;
    if (pep[1]<>pep[x]) or (x=2) then perm(pep,x-1);
  end
  else
  begin
    score:=0;
    hitp:=0;
    makepeak(pep);
    for k:=1 to nspec do
    begin
      score:=score+calcs(speak[k]);
    end;
    if ((hitp+1>minp) and (score<maxs) and (score/hitp<maxp))
    then makerep(pep);               {make report}
  end;
end;

{*****************************}
Procedure sorted(var what:s50);             {sort peptide sequence (Qsort)}
{*****************************}

Procedure Partsort(a,b : byte; var s:s50);
var i,j,k : byte;

Procedure Switch(var a,b : char);
var c : char;
begin
  if a<>b then
  begin
    c:=a;
    a:=b;
    b:=c;
  end;
end;

begin
  k:=(ord(s[a])+ord(s[b])) div 2;
  i:=a;
  j:=b;
  repeat
    while ord(s[i])<k do inc(i,1);
    while k<ord(s[j]) do dec(j,1);
    if i<=j then
    begin
      Switch(s[i],s[j]);
      inc(i,1);
      dec(j,1);
    end;
  until i>j;
  if a<j then Partsort(a,j,s);
  if i<b then Partsort(i,b,s);
end;

begin
  partsort(1,ord(what[0]),what);
end;

{************************************}
Procedure calcw(var mw:real; aa:byte);      {calcutate mol weight}
{************************************}

var i,j,k,l,n : byte;
    m         : real;
    pep       : s50;

begin
  j:=trunc(mw/amiw[amic[aa]])+1;       {maximum no of aa that fits in mw}
  if j>maxa then j:=maxa;
  for i:=0 to j do
  begin
    amix[aa]:=i;
    m:=mw-i*amiw[amic[aa]];            {calculate remaining mw and go recursive}
    if (aa>1) and (m>-1*tole) then calcw(m,aa-1)
    else
    begin
      if keypressed then stopit;
      molc:=0;
      l:=0;
      for k:=1 to amir do
      begin
        molc:=molc+amix[k]*amiw[amic[k]];
        l:=l+amix[k];
      end;
      molc:=molc+cend+nend;
      inc(check);
      gotoxy(1,2);write(check:20);
      gotoxy(1,6);for k:=1 to amir do write(amix[k]:3);
      if (abs(molc-molw)<tole) and (l-mina+1>0) and (maxa-l+1>0) then
      begin
        n:=0;
        for k:=1 to amir do
        begin
         if amix[k]<>0 then for l:=1 to amix[k] do
          begin
            inc(n);
            pep[n]:=char(amic[k]);
          end;
        end;
        pep[0]:=chr(n);
        if minp=0 then
        begin
          makepeak(pep);
          makerep(pep);
        end
        else 
        begin
          sorted(pep);
          perm(pep,n);                 {make permutations and report}
        end
      end;
    end;
  end;
end;

{***************}
Procedure setdef;                      {set default values & initialize}
{***************}

begin
nend:=H;
cend:=OH;
maxs:=9999;
maxp:=9999;
maxh:=999999;
maxa:=50;
mina:=1;
tole:=1;
tolp:=99;
minp:=0;
molw:=500;
score:=0;
hitp:=0;
check:=0;
hit:=0;
repo:='#QMDS/PC';
ions:='ABCXYZ';
titl:='PEPSEQ CALCULATION OUTPUT';
rot:=60;
end;

{****************}
Procedure readinp;
{****************}

var i  : byte;
    f  : text;
    a5 : s5;
    a1 : s1;
    
begin
  if paramcount=1 then 
    inpf:=paramstr(1)
  else inpf:='PEPSEQ.INP';
  assign(f,inpf);
  reset(f);
  if ioresult<>0 then
  begin
    writeln(#7,'File ',inpf,' not found.');
    halt;
  end;
  nspec:=0;
  datf:='';
  while not eof(f) do
  begin
    read(f,a5);
    if a5='TITL=' then readln(f,titl)
    else if a5='DATA=' then readln(f,datf)
    else if a5='CEND=' then readln(f,cend)
    else if a5='NEND=' then readln(f,nend)
    else if a5='REPO=' then readln(f,repo)
    else if a5='MAXA=' then readln(f,maxa)
    else if a5='MINA=' then readln(f,mina)
    else if a5='TOLE=' then readln(f,tole)
    else if a5='TOLP=' then readln(f,tolp)
    else if a5='MINP=' then readln(f,minp)
    else if a5='MAXS=' then readln(f,maxs)
    else if a5='MAXP=' then readln(f,maxp)
    else if a5='MAXH=' then readln(f,maxh)
    else if a5='MOLW=' then readln(f,molw)
    else if a5='IONS=' then readln(f,ions)
    else if a5='MOLH=' then 
    begin
      readln(f,molw);
      molw:=molw-H;
    end
    else if a5='PEAK=' then
    begin
      inc(nspec);
      readln(f,speak[nspec]);
    end
    else readln(f);
  end;
  close(f);
  if datf='' then datf:='PEPSEQ.DAT';
  assign(f,datf);
  reset(f);
  if ioresult<>0 then
  begin
    writeln(#7,'File ',datf,' not found.');
    halt;
  end;
  amir:=0;
  gotoxy(1,1);writeln('#checked':20,'hits':10,'!':2);
  gotoxy(1,4);
  while (amir<255) and (not eof(f)) do
  begin
    inc(amir);
    amix[amir]:=0;
    read(f,a1);
    amic[amir]:=ord(a1[1]);
    read(f,amiw[amic[amir]]);
    readln(f);
    write(a1:3);
  end;
  close(f);
end;

begin
clrscr;
setdef;
readinp;
assign(outf,'PEPSEQ.OUT');
rewrite(outf);
rephead;
temp:=molw-nend-cend;
calcw(temp,amir);
stopit;
end.
